local cql = ClassicQuestLog

--[[
    Large sections of code here are shamelessly copied from Blizzard's FrameXML\QuestInfo.lua (and xml)
    with minor tweaks like removing global names that would collide with Blizzard's. This allows this
    addon to have its own detail frame without borrowing the default's frame
]]

-- this is QUEST_TEMPLATE_LOG but only elements are used here
local template = { questLog = true, chooseItems = nil, contentWidth = 285,
	canHaveSealMaterial = true, sealXOffset = 160, sealYOffset = -6,
	elements = {
		{"ShowTitle", 0, 0},
		{"ShowStatusTitle", 0, -10},
		{"ShowStatusText", 0, -5},
		{"ShowTypeFrame", 0, -8},
		{"ShowObjectivesText", 0, -5},
		{"ShowTimer", 0, -10},
		{"ShowObjectives", 0, -10},
		{"ShowSpecialObjectives", 0, -10},
		{"ShowRequiredMoney", 0, 0},
		{"ShowGroupSize", 0, -10},
		{"ShowDescriptionHeader", 0, -20},
		{"ShowDescriptionText", 0, -5},
		{"ShowSeal", 0, 0},
		{"ShowRewards", 0, -10},
		{"ShowSpacer", 0, -10}
	}
}

local DEFAULT_TEXT_COLOR = CreateColor(0,0,0,1)
local COMPLETED_TEXT_COLOR = CreateColor(0.2,0.2,0.2,1)
local RED_TEXT_COLOR = CreateColor(0.5,0,0,1)

local REWARDS_SECTION_OFFSET = 5		-- vertical distance between sections
local REWARDS_ROW_OFFSET = 4			-- vertical distance between rows within a section
local REWARDS_BUTTON_WIDTH = 149        -- width of the reward buttons

local QUEST_INFO_SPELL_REWARD_ORDERING = {
	Enum.QuestCompleteSpellType.Follower,
	Enum.QuestCompleteSpellType.Companion,
	Enum.QuestCompleteSpellType.Tradeskill,
	Enum.QuestCompleteSpellType.Ability,
	Enum.QuestCompleteSpellType.Aura,
	Enum.QuestCompleteSpellType.Spell,
	Enum.QuestCompleteSpellType.Unlock,
}

local QUEST_INFO_SPELL_REWARD_TO_HEADER = {
	[Enum.QuestCompleteSpellType.Follower] = REWARD_FOLLOWER,
	[Enum.QuestCompleteSpellType.Companion] = REWARD_COMPANION,
	[Enum.QuestCompleteSpellType.Tradeskill] = REWARD_TRADESKILL_SPELL,
	[Enum.QuestCompleteSpellType.Ability] = REWARD_ABILITY,
	[Enum.QuestCompleteSpellType.Aura] = REWARD_AURA,
	[Enum.QuestCompleteSpellType.Spell] = REWARD_SPELL,
	[Enum.QuestCompleteSpellType.Unlock] = REWARD_UNLOCK,
}

function cql.Detail:Setup()
    self:SetScript("OnSizeChanged",self.OnSizeChanged)
    self:OnSizeChanged(self:GetWidth(),self:GetHeight())
end

function cql.Detail:OnSizeChanged(width,height)
    local content = self.ScrollFrame.Content
    local contentWidth = max(floor(width+0.5)-35,150)
    content:SetWidth(contentWidth)
    content.width = contentWidth

    self:Update() -- necessary for detail elements to adjust
end

function cql.Detail.ScrollFrame.Content.PortraitFrame:OnEnter()
	if self.portraitName or self.portraitText then
		GameTooltip:SetOwner(self,"ANCHOR_LEFT")
		if self.portraitName and self.portraitName:len()>0 then
			GameTooltip:SetText(self.portraitName,1,0.82,0,1)
		end
		if self.portraitText and self.portraitText:len()>0 then
			GameTooltip:AddLine(self.portraitText,1,1,1,1)
		end
		GameTooltip:Show()
	end
end

function cql.Detail.ScrollFrame.Content.PortraitFrame:OnLeave()
	GameTooltip:Hide()
end

function cql.Detail:Update()
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()

	if not self:IsVisible() then
		return
	end

	-- details background
	local theme = C_QuestLog.GetQuestDetailsTheme(questID)
	content.SealFrame.theme = theme

	if questID==0 then
		self.Back:SetAtlas("NoQuestsBackground")
		self.Back:SetVertexColor(1,1,1)
	elseif cql.settings.HighContrast then
		self.Back:SetAtlas("QuestDetailsBackgrounds-Accessibility")
		self.Back:SetVertexColor(0.8,0.75,0.65)
	elseif theme and theme.background and not cql.settings.HideCampaignProgress then
		self.Back:SetAtlas(theme.background)
		self.Back:SetVertexColor(1,1,1)
	else
		self.Back:SetAtlas("QuestDetailsBackgrounds")
		self.Back:SetVertexColor(1,1,1)
	end

	self.Back:SetDesaturated(C_QuestLog.IsQuestDisabledForSession(questID))

	local questPortrait, questPortraitText, questPortraitName, questPortraitMount, questPortraitModelSceneID = C_QuestLog.GetQuestLogPortraitGiver()
	local portraitFrame = self.ScrollFrame.Content.PortraitFrame
	local portraitHeight = 0
	if questID>0 and (questPortrait and questPortrait ~= 0 and QuestLogShouldShowPortrait()) then
		if cql.settings.ShowDefaultQuestPortrait then
			QuestFrame_ShowQuestPortrait(ClassicQuestLog, questPortrait, questPortraitMount, questPortraitModelSceneID, questPortraitText, questPortraitName, -2, -43)
			portraitFrame:Hide()
		else
			SetPortraitTextureFromCreatureDisplayID(portraitFrame.Portrait,questPortrait)
			portraitFrame.portraitName = questPortraitName
			portraitFrame.portraitText = questPortraitText
			portraitHeight = ceil(portraitFrame:GetHeight())
			portraitFrame:Show()
		end
	else
		if cql.settings.ShowDefaultQuestPortrait then
			QuestFrame_HideQuestPortrait()
		end
		portraitFrame:Hide()
	end

	-- show elements of details frame by calling their template function
	local lastFrame
	local height = 0
	for i,info in ipairs(template.elements) do
		if self[info[1]] then
			local shownFrame,bottomShownFrame = self[info[1]](self,height<portraitHeight and portraitHeight or 0)
			if shownFrame then
--				shownFrame:SetParent(content)
				shownFrame:ClearAllPoints()
				if lastFrame then
					shownFrame:SetPoint("TOPLEFT",lastFrame,"BOTTOMLEFT",info[2],info[3])
				else
					shownFrame:SetPoint("TOPLEFT",content,"TOPLEFT",info[2],info[3])
				end
				lastFrame = bottomShownFrame or shownFrame
				height = height + lastFrame:GetHeight()
			end
		end
	end
end

-- title is always shown
function cql.Detail:ShowTitle(padding)
    local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()

	if questID==0 then
		content.TitleHeader:Hide()
		return
	end

    local title = C_QuestLog.GetTitleForQuestID(questID)
    if IsCurrentQuestFailed() then
		title = QUEST_TITLE_FORMAT_FAILED:format(title)
	end

	content.TitleHeader:Show()
    content.TitleHeader:SetText(title)
	content.TitleHeader:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
    content.TitleHeader:SetWidth(floor(content:GetWidth()+0.5)-padding)

	return content.TitleHeader
end

-- this is not in default; used for showing extra status title (at the moment quest disabled due to party sync)
-- the title and text are separate fontstrings to make it easier to resize
function cql.Detail:ShowStatusTitle(padding)
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()

	if questID>0 and C_QuestLog.IsQuestDisabledForSession(questID) then
		content.StatusTitle:Show()
		content.StatusTitle:SetText("\124A:QuestSharing-QuestDetails-Padlock:16:16\124a"..QUEST_SESSION_ON_HOLD_TOOLTIP_TITLE)
		content.StatusTitle:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		content.StatusTitle:SetWidth(floor(content:GetWidth()+0.5)-padding)
		return content.StatusTitle
	else
		content.StatusTitle:Hide()
	end
end

-- this is not in default; goes with StatusTitle to show text description of status
function cql.Detail:ShowStatusText(padding)
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()
	local text

	if questID>0 and C_QuestLog.IsQuestDisabledForSession(questID) then
		content.StatusText:Show()
		content.StatusText:SetText(QUEST_SESSION_ON_HOLD_TOOLTIP_TEXT.."\n ")
		content.StatusText:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		content.StatusText:SetWidth(floor(content:GetWidth()+0.5)-padding)
		return content.StatusText
	else
		content.StatusText:Hide()
	end
end

-- replaces ShowType; instead of showing a fontstring of an icon and name, it shows a set of type frames
-- to display tag, faction(for account-wide quests if faction specific) and frequency
function cql.Detail:ShowTypeFrame(padding)
    local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()
	local totalWidth = floor(content:GetWidth()+0.5)-padding

	content.TypeFrame:SetWidth(totalWidth)

	for _,button in ipairs(content.TypeFrame.Buttons) do
		button:Hide()
	end

	local index,xoff,yoff = 1,0,0

	local function addType(icon,text)
		local button = self.ScrollFrame.Content.TypeFrame.Buttons[index]
		if not button then
			button = CreateFrame("Frame",nil,self.ScrollFrame.Content.TypeFrame,"ClassicQuestLogQuestTypeTemplate")
			self.ScrollFrame.Content.TypeFrame.Buttons[index] = button
		end
		button.Icon:SetAtlas(icon)
		button.Text:SetText(text)
		button.Text:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		for _,outline in ipairs(button.Outlines) do
			outline:SetAtlas(icon)
			outline:SetVertexColor(0,0,0,0.5)
		end
		button.width = 20+2+ceil(button.Text:GetStringWidth())
		button:SetWidth(button.width)
		if xoff + button.width > totalWidth then
			xoff = 0
			yoff = yoff - 18
		end
		button:ClearAllPoints()
		button:SetPoint("TOPLEFT",xoff,yoff)
		button:Show()
		xoff = xoff + button.width
		index = index + 1
		return button
	end

	-- primary quest tag goes first (this one can use non-dungeon tags, such as Account)
	local info = C_QuestLog.GetQuestTagInfo(questID)
	if info then
		local isDungeonType = (info.worldQuestType and WORLD_QUEST_TYPE_ATLAS[info.worldQuestType] or (info.tagID and QUEST_TAG_DUNGEON_TYPES[info.tagID])) and true or false
		if isDungeonType or cql.settings.ShowExtraTypeDetails then
			local atlas = info.worldQuestType and WORLD_QUEST_TYPE_ATLAS[info.worldQuestType] or (info.tagID and QUEST_TAG_ATLAS[info.tagID])
			if atlas then
				addType(atlas,info.tagName)
			end
		end
	end

	-- if 'Show Extra Quest Details' is enabled, show other icon+text for account/daily/etc
	if cql.settings.ShowExtraTypeDetails then
		-- faction if needed
		local faction = GetQuestFactionGroup(questID)
		if faction and info and info.tagID==Enum.QuestTag.Account then
			if faction==LE_QUEST_FACTION_ALLIANCE then
				addType(QUEST_TAG_ATLAS["ALLIANCE"],FACTION_ALLIANCE)
			elseif faction==LE_QUEST_FACTION_HORDE then
				addType(QUEST_TAG_ATLAS["HORDE"],FACTION_HORDE)
			end
		end

		--  daily or weekly
		local index = C_QuestLog.GetLogIndexForQuestID(questID)
		if index then
			local info = C_QuestLog.GetInfo(index)
			if info and info.frequency==Enum.QuestFrequency.Daily then
				addType(QUEST_TAG_ATLAS["DAILY"],DAILY)
			elseif info and info.frequency==Enum.QuestFrequency.Weekly then
				addType(QUEST_TAG_ATLAS["WEEKLY"],WEEKLY)
			end
		end

		-- complete or failed
		if C_QuestLog.IsComplete(questID) then
			addType(QUEST_TAG_ATLAS["COMPLETED"],COMPLETE)
		elseif C_QuestLog.IsFailed(questID) then
			addType(QUEST_TAG_ATLAS["FAILED"],FAILED)
		end
	end

	if xoff>0 or yoff<0 then
		content.TypeFrame:SetHeight(-yoff+18)
		content.TypeFrame:Show()
		return content.TypeFrame
	else
		content.TypeFrame:Hide()
	end
end

-- short summary of quest always shown
function cql.Detail:ShowObjectivesText(padding)
    local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()

	if questID~=0 then
		local _,questObjectives = GetQuestLogQuestText()
		content.ObjectivesText:Show()
		content.ObjectivesText:SetText(questObjectives)
		content.ObjectivesText:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		content.ObjectivesText:SetWidth(floor(content:GetWidth()+0.5)-padding)
		return content.ObjectivesText
	else
		content.ObjectivesText:Hide()
	end

end

-- for ShowObjectives, returns an indexed fontstring for ObjectivesFrame, creating if needed
function cql.Detail:AcquireObjective(index)
	local content = self.ScrollFrame.Content

	local newObjective = content.ObjectivesFrame.Objectives[index]
	if not newObjective then
		newObjective = content.ObjectivesFrame:CreateFontString(nil,"BACKGROUND","QuestFontNormalSmall")
		newObjective:SetPoint("TOPLEFT",content.ObjectivesFrame.Objectives[index-1],"BOTTOMLEFT",0,-2)
		newObjective:SetJustifyH("LEFT")
		newObjective:SetWidth(285)
		content.ObjectivesFrame.Objectives[index] = newObjective
	end

	return newObjective
end

-- a frame that contains a variable number of fontstrings to specify progress
function cql.Detail:ShowObjectives(padding)
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()
	local numObjectives = GetNumQuestLeaderBoards()
	local objective
	local text, type, finished
	local objectivesTable = content.ObjectivesFrame.Objectives
	local numVisibleObjectives = 0

	if questID==0 then
		content.ObjectivesFrame:Hide()
		return
	end

	local waypointText = C_QuestLog.GetNextWaypointText(questID)
	if waypointText then
		numVisibleObjectives = numVisibleObjectives + 1
		objective = self:AcquireObjective(numVisibleObjectives)
		objective:SetText(WAYPOINT_OBJECTIVE_FORMAT_OPTIONAL:format(waypointText))
		objective:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		objective:SetWidth(floor(content:GetWidth()+0.5)-padding)
		objective:Show()
	end

	for i=1,numObjectives do
		text,type,finished = GetQuestLogLeaderBoard(i)
		if (type~="spell" and type~="log" and numVisibleObjectives<MAX_OBJECTIVES) then
			numVisibleObjectives = numVisibleObjectives + 1
			objective = self:AcquireObjective(numVisibleObjectives)
			if not text or text:len()==0 then
				text = type
			end
			if finished then
				objective:SetTextColor(COMPLETED_TEXT_COLOR:GetRGB())
				text = text.." ("..COMPLETE..")"
			else
				objective:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
			end
			objective:SetText(text)
			objective:SetWidth(floor(content:GetWidth()+0.5)-padding)
			objective:Show()
		end
	end

	for i=numVisibleObjectives+1,#objectivesTable do
		objectivesTable[i]:Hide()
	end

	if objective then
		content.ObjectivesFrame:Show()
		return content.ObjectivesFrame, objective
	else
		content.ObjectivesFrame:Hide()
	end
end

-- a frame that shows a spell objective (Learn Spell: and button for spell to learn)
function cql.Detail:ShowSpecialObjectives()
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()

	local spellID, spellName, spellTexture, finished = GetQuestLogCriteriaSpell()
	--local spellID, spellName, spellTexture, finished = 132621, "Teleport: Vale of Eternal Blossoms", 851298, true

	local spellObjectiveFrame = content.SpecialObjectivesFrame.SpellObjectiveFrame
	local spellObjectiveLearnLabel = content.SpecialObjectivesFrame.SpellObjectiveLearnLabel

	if spellID and questID~=0 then
		spellObjectiveFrame.Icon:SetTexture(spellTexture)
		spellObjectiveFrame.Name:SetText(spellName)
		spellObjectiveFrame.spellID = spellID

		spellObjectiveFrame:ClearAllPoints()
		spellObjectiveLearnLabel:SetPoint("TOPLEFT",0,0)
		spellObjectiveFrame:SetPoint("TOPLEFT",spellObjectiveLearnLabel,"BOTTOMLEFT",0,-4)

		if finished then
			spellObjectiveLearnLabel:SetText(LEARN_SPELL_OBJECTIVE.." ("..COMPLETE..")")
			spellObjectiveLearnLabel:SetTextColor(COMPLETED_TEXT_COLOR:GetRGB())
		else
			spellObjectiveLearnLabel:SetText(LEARN_SPELL_OBJECTIVE)
			spellObjectiveLearnLabel:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		end

		spellObjectiveLearnLabel:Show()
		spellObjectiveFrame:Show()

		content.SpecialObjectivesFrame:SetHeight(spellObjectiveFrame:GetHeight() + spellObjectiveLearnLabel:GetHeight())
		content.SpecialObjectivesFrame:Show()
		return content.SpecialObjectivesFrame
	else
		spellObjectiveFrame:Hide()
		spellObjectiveLearnLabel:Hide()
		content.SpecialObjectivesFrame:Hide()
	end
end

-- a frame that shows a money objective (Required Money:)
function cql.Detail:ShowRequiredMoney()
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()

	if questID==0 then
		content.RequiredMoneyFrame:Hide()
		return
	end

	local requiredMoney = C_QuestLog.GetRequiredMoney()

	if requiredMoney > 0 then
		MoneyFrame_Update(content.RequiredMoneyFrame.RequiredMoneyDisplay,requiredMoney)
		if requiredMoney > GetMoney() then
			content.RequiredMoneyFrame.RequiredMoneyText:SetTextColor(RED_TEXT_COLOR:GetRGB())
		else
			content.RequiredMoneyFrame.RequiredMoneyText:SetTextColor(COMPLETED_TEXT_COLOR:GetRGB())
		end
		content.RequiredMoneyFrame:Show()
		return content.RequiredMoneyFrame
	else
		content.RequiredMoneyFrame:Hide()
	end
end

-- text to show suggested group size
function cql.Detail:ShowGroupSize()
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()
	local groupNum = C_QuestLog.GetSuggestedGroupSize(questID)

	if questID>0 and groupNum>0 then
		content.GroupSize:SetText(format(QUEST_SUGGESTED_GROUP_NUM, groupNum))
		content.GroupSize:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		content.GroupSize:Show()
		return content.GroupSize
	else
		content.GroupSize:Hide()
	end
end

-- always-shown big Description header
function cql.Detail:ShowDescriptionHeader()
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()
	if questID>0 then
		content.DescriptionHeader:Show()
		return content.DescriptionHeader
	else
		content.DescriptionHeader:Hide()
	end
end

-- always-shown long description of the quest
function cql.Detail:ShowDescriptionText()
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()

	if questID>0 then
		local questDescription = GetQuestLogQuestText()
		content.DescriptionText:Show()
		content.DescriptionText:SetText(questDescription)
		content.DescriptionText:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		content.DescriptionText:SetWidth(floor(content:GetWidth()+0.5))
		return content.DescriptionText
	else
		content.DescriptionText:Hide()
	end
end

-- 5px spacer at very end of details so text doesn't sit at bottom pixel of scrollframe
function cql.Detail:ShowSpacer()
	return self.ScrollFrame.Content.SpacerFrame
end

-- a large badge and "signature" for some quests
function cql.Detail:ShowSeal()
	local content = self.ScrollFrame.Content
	local questID = cql.Log:GetSelectedQuest()
	local frame = content.SealFrame

	if questID==0 then
		frame:Hide()
		return
	end

	local theme = frame.theme
	local hasAnyPartOfTheSeal = theme and (theme.signature ~= "" or theme.seal)
	frame:SetShown(hasAnyPartOfTheSeal)

	if hasAnyPartOfTheSeal then
		frame:SetPoint("CENTER",content)
		frame.Text:SetText(theme.signature)
		frame.Text:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		frame.Texture:SetShown(theme.seal~=nil)
		if theme.seal then
			frame.Texture:SetAtlas(theme.seal,true)
		end
		frame:SetWidth(floor(content:GetWidth()+0.5))
		return frame
	end
end

function cql.Detail:GetRewardSpellBucketType(spellInfo)
	if spellInfo.type and spellInfo.type ~= Enum.QuestCompleteSpellType.LegacyBehavior then
		return spellInfo.type
	elseif spellInfo.isTradeskillSpell then
		return Enum.QuestCompleteSpellType.Tradeskill
	elseif spellInfo.isBoostSpell then
		return Enum.QuestCompleteSpellType.Ability
	elseif spellInfo.garrFollowerID then
		local followerInfo = C_Garrison.GetFollowerInfo(spellInfo.garrFollowerID)
		if followerInfo and followerInfo.followerTypeID == Enum.GarrisonFollowerType.FollowerType_9_0_GarrisonFollower then
			return Enum.QuestCompleteSpellType.Companion
		else
			return Enum.QuestCompleteSpellType.Follower
		end
	elseif spellInfo.isSpellLearned then
		return Enum.QuestCompleteSpellType.Spell
	elseif spellInfo.genericUnlock then
		return Enum.QuestCompleteSpellType.Unlock
	end

	return Enum.QuestCompleteSpellType.Aura
end


--[[ helper functions for AddFontStringToRewards and AddRewardButtonToRewards ]]

local fontStringPool = {}

-- get an available fontstring from the pool or creates if needed
function cql.Detail:AcquireFontString()
	local rewardsFrame = cql.Detail.ScrollFrame.Content.RewardsFrame
	for _,fontstring in ipairs(fontStringPool) do
		if not fontstring.isUsed then
			fontstring.isUsed = true
			fontstring:Show()
			return fontstring
		end
	end
	local fontstring = rewardsFrame:CreateFontString(nil,"ARTWORK","QuestFont")
	fontstring:SetJustifyH("LEFT")
	fontstring.isUsed = true
	tinsert(fontStringPool,fontstring)
	return fontstring
end

function cql.Detail:ReleaseFontStrings()
	for _,fontstring in ipairs(fontStringPool) do
		fontstring.isUsed = false
		fontstring:ClearAllPoints()
		fontstring:Hide()
	end
end

local rewardButtonPool = {}

-- gets an available reward button from the pool or creates if needed
function cql.Detail:AcquireRewardButton()
	for _,button in ipairs(rewardButtonPool) do
		if not button.isUsed then
			button.isUsed = true
			button:SetWidth(REWARDS_BUTTON_WIDTH)
			button:Show()
			return button
		end
	end
	-- if still here, need to make a new one
	local button = CreateFrame("Button",nil,cql.Detail.ScrollFrame.Content,"ClassicQuestLogRewardButtonTemplate")
	button.isUsed = true
	button:SetWidth(REWARDS_BUTTON_WIDTH)
	tinsert(rewardButtonPool,button)
	return button
end

-- releases all reward buttons in the pool
function cql.Detail:ReleaseRewardButtons()
	for _,button in ipairs(rewardButtonPool) do
		button.isUsed = false
		button:ClearAllPoints()
		button:Hide()
	end
end

-- updates a rewardbutton to the given arguments
function cql.Detail:FillReward(rewardButton,questID,rewardType,rewardIndex,objectType,id,name,icon,quality,amount)
	rewardButton.questID = questID
	rewardButton.rewardType = rewardType
	rewardButton.rewardIndex = rewardIndex
	rewardButton.objectType = objectType
	rewardButton.id = id
	rewardButton.Text:SetText(name)
	rewardButton.Icon:SetTexture(icon)
	if type(quality)=="table" and #quality==3 then -- if r,g,b passed for quality use this color
		rewardButton.Rarity:SetVertexColor(unpack(quality))
	elseif quality and BAG_ITEM_QUALITY_COLORS[quality] then
		local color = BAG_ITEM_QUALITY_COLORS[quality]
		rewardButton.Rarity:SetVertexColor(color.r,color.g,color.b)
	else
		rewardButton.Rarity:SetVertexColor(0.5,0.5,0.5)
	end
	if amount and (type(amount)~="number" or amount>1) then
		rewardButton.Amount:SetText(amount)
		rewardButton.Amount:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
		rewardButton.Amount:Show()
	else
		rewardButton.Amount:Hide()
	end
	rewardButton.Cosmetic:SetShown(objectType=="item" and IsCosmeticItem(id))
	-- if filling an item that's not cached, come back and try again
	if objectType=="item" and not name or name=="" then
		cql.timer:Start(0.5,cql.DelayedUpdate)
	end
end

-- anchors an object to the lastAnchorElement (padding based on isSection) and makes object new lastAnchorElement
function cql.Detail:AnchorReward(object,isSection)
	local rewardsFrame = cql.Detail.ScrollFrame.Content.RewardsFrame
	local offset = isSection and -REWARDS_SECTION_OFFSET or -REWARDS_ROW_OFFSET
	local height = object:GetObjectType()=="FontString" and object:GetStringHeight() or object:GetHeight()
	if isSection then
		rewardsFrame.xoff = 0 -- if starting a new section, then rewardButtons should start over from left
	end
	object:ClearAllPoints()
	object:SetPoint("TOPLEFT",rewardsFrame.lastAnchorElement,"BOTTOMLEFT",0,offset)
	object:Show()
	rewardsFrame.totalHeight = rewardsFrame.totalHeight + height + offset
	rewardsFrame.lastAnchorElement = object
end

-- adds a fontstring with given text to rewards (need to set width before setting text to get proper string height)
function cql.Detail:AddTextToRewards(fontstring,text)
	local rewardsFrame = cql.Detail.ScrollFrame.Content.RewardsFrame
	fontstring:SetWidth(floor(rewardsFrame:GetWidth()+0.5))
	fontstring:SetText(text)
	fontstring:SetTextColor(DEFAULT_TEXT_COLOR:GetRGB())
	cql.Detail:AnchorReward(fontstring,true)
end

-- anchors the rewardButton either to left under rewardsFrame.lastAnchorElement or to right of last rewardButton if space permits
-- when a section will have only one rewardButton, then a anchorReward by itself is fine
function cql.Detail:AnchorRewardButton(rewardButton)
	local rewardsFrame = cql.Detail.ScrollFrame.Content.RewardsFrame
	local contentWidth = floor(rewardsFrame:GetWidth()+0.5)
	if rewardsFrame.xoff==0 then
		cql.Detail:AnchorReward(rewardButton)
		rewardsFrame.xoff = rewardsFrame.xoff + REWARDS_BUTTON_WIDTH + 4
		if rewardsFrame.xoff+REWARDS_BUTTON_WIDTH > contentWidth then
			rewardsFrame.xoff = 0
		end
	else
		rewardButton:ClearAllPoints()
		rewardButton:SetPoint("TOPLEFT",rewardsFrame.lastAnchorElement,"TOPLEFT",rewardsFrame.xoff,0)
		rewardsFrame.xoff = rewardsFrame.xoff + REWARDS_BUTTON_WIDTH + 4
		if rewardsFrame.xoff+REWARDS_BUTTON_WIDTH > contentWidth then
			rewardsFrame.xoff = 0
		end
	end
end

--[[ these should be used to add all rewards, either in button or text form ]]

-- these are the icon+text buttons that can stack horizontally (if space permits) as well as vertically
function cql.Detail:AddRewardButtonToRewards(questID,rewardType,rewardIndex,objectType,id,name,icon,quality,amount)
	local rewardButton = cql.Detail:AcquireRewardButton()
	cql.Detail:AnchorRewardButton(rewardButton)
	cql.Detail:FillReward(rewardButton,questID,rewardType,rewardIndex,objectType,id,name,icon,quality,amount)
	return rewardButton
end

-- these are for section headers and other text that uses QuestFont that can wrap to window
function cql.Detail:AddFontStringToRewards(text)
	local fontstring = cql.Detail:AcquireFontString()
	cql.Detail:AddTextToRewards(fontstring,text)
	return fontstring
end

--[[ the main RewardsFrame update ]]

function cql.Detail:ShowRewards()
	local content = self.ScrollFrame.Content
	local rewardsFrame = content.RewardsFrame
	local questID = cql.Log:GetSelectedQuest()

	rewardsFrame.totalHeight = rewardsFrame.Header:GetStringHeight() -- running total of rewards height
	rewardsFrame.lastAnchorElement = rewardsFrame.Header -- Rewards header is always present if anything else is
	rewardsFrame.xoff = 0 -- xoffset for rewards buttons, resets to 0 for seconds

	local contentWidth = floor(content:GetWidth()+0.5)
	rewardsFrame:SetWidth(contentWidth)

	cql.Detail:ReleaseRewardButtons()
	cql.Detail:ReleaseFontStrings()

	if questID==0 then
		rewardsFrame:Hide()
		return
	end

	-----[[ gathering rewards data ]]-----

	-- gather rewards data
	local numQuestRewards = GetNumQuestLogRewards()
	local numQuestChoices = GetNumQuestLogChoices(questID, true)
	local numQuestCurrencies = GetNumQuestLogRewardCurrencies()
	local money = GetQuestLogRewardMoney()
	local skillName, skillIcon, skillPoints = GetQuestLogRewardSkillPoints()
	local xp = GetQuestLogRewardXP()
	local artifactXP, artifactCategory = GetQuestLogRewardArtifactXP()
	local honor = GetQuestLogRewardHonor()
	local playerTitle = GetQuestLogRewardTitle()
	local spellRewards = C_QuestInfoSystem.GetQuestRewardSpells(questID) or {}
	local spellRewardBuckets = {}
	ProcessQuestLogRewardFactions()
	local hasWarModeBonus = C_QuestLog.QuestHasWarModeBonus(questID)
	local majorFactionRepRewards = C_QuestLog.GetQuestLogMajorFactionReputationRewards(questID)
	local hasChanceForQuestSessionBonusReward = C_QuestLog.QuestHasQuestSessionBonus(questID)

	-- gather spells to be rewarded (can be garrison followers, tradeskills, abilities) into spellRewardBuckets
	for index, spellID in ipairs(spellRewards) do
		if spellID and spellID > 0 then
			local spellInfo = C_QuestInfoSystem.GetQuestRewardSpellInfo(questID, spellID)
			local knownSpell = IsSpellKnownOrOverridesKnown(spellID)
			-- only allow the spell reward if user can learn it
			if spellInfo and spellInfo.texture and not knownSpell and (not spellInfo.isBoostSpell or IsCharacterNewlyBoosted()) and (not spellInfo.garrFollowerID or not C_Garrison.IsFollowerCollected(spellInfo.garrFollowerID)) then
				spellInfo.spellID = spellID
				local subType = cql.Detail:GetRewardSpellBucketType(spellInfo)
				if not spellRewardBuckets[subType] then
					spellRewardBuckets[subType] = {}
				end
				tinsert(spellRewardBuckets[subType],spellInfo)
			end
		end
	end

	-- if no rewards to display, leave
	local totalRewards = numQuestRewards + numQuestChoices + numQuestCurrencies
	if totalRewards == 0 and money == 0 and xp == 0 and not playerTitle and #spellRewards == 0 and artifactXP == 0 and honor == 0 and not majorFactionRepRewards then
		rewardsFrame:Hide()
		return nil
	end

	-----[[ placing rewards begins here ]]-----

	-- choosable rewards (items, currency)
	if numQuestChoices > 0 then
		-- "You will receive:" or "You will be able to choose one of these rewards:"
		cql.Detail:AddFontStringToRewards(numQuestChoices==1 and REWARD_ITEMS_ONLY or REWARD_CHOICES)

		for i=1,numQuestChoices do
			local lootType = GetQuestLogChoiceInfoLootType(i)
			if lootType==0 then -- LOOT_LIST_ITEM
				local name, texture, numItems, quality, isUsable, itemID = GetQuestLogChoiceInfo(i)
				cql.Detail:AddRewardButtonToRewards(questID,"choice",i,"item",itemID,name,texture,quality,numItems)
			elseif lootType==1 then -- LOOT_LIST_CURRENCY
				local name, texture, amount, currencyID, quality = GetQuestLogRewardCurrencyInfo(i,questID,true) -- true for isChoice
				local info = C_CurrencyInfo.GetCurrencyContainerInfo(currencyID,amount)
				if info then -- if GetCurrencyContainerInfo changes any details
					cql.Detail:AddRewardButtonToRewards(questID,"choice",i,"currency",{currencyID=currencyID,amount=amount},info.name,info.icon,info.quality,info.displayAmount)
				else
					cql.Detail:AddRewardButtonToRewards(questID,"choice",i,"currency",{currencyID=currencyID,amount=amount},name,texture,quality,amount)
				end

			end
		end

	end

	-- spell rewards (includes followers, tradeskills, abilities, buffs and spells)
	if #spellRewards > 0 then
		for orderIndex,bucketType in ipairs(QUEST_INFO_SPELL_REWARD_ORDERING) do
			local bucket = spellRewardBuckets[bucketType]
			if bucket then
				for i,spellInfo in ipairs(bucket) do
					if i==1 and not spellInfo.hideSpellLearnText then
						cql.Detail:AddFontStringToRewards(QUEST_INFO_SPELL_REWARD_TO_HEADER[bucketType])
					end
					if spellInfo.garrFollowerID then -- garrison follower
						local followerInfo = C_Garrison.GetFollowerInfo(spellInfo.garrFollowerID)
						cql.Detail:AddRewardButtonToRewards(questID,"spells",i,"follower",spellInfo.garrFollowerID,followerInfo.name,followerInfo.portraitIconID or "Interface\\Garrison\\Portraits\\FollowerPortrait_NoPortrait",followerInfo.quality)
					else -- regular spell
						cql.Detail:AddRewardButtonToRewards(questID,"spells",i,"spell",spellInfo.spellID,spellInfo.name,spellInfo.texture,{0.75,0.615,0})
					end
				end
			end
		end
	end

	-- title reward
	if playerTitle then
		-- "You shall be granted the title:"
		cql.Detail:AddFontStringToRewards(REWARD_TITLE)
		local rewardButton = cql.Detail:AddRewardButtonToRewards(questID,"title",nil,"title",nil,playerTitle,"Interface\\Icons\\INV_Misc_Note_02")
		rewardButton:SetWidth(contentWidth)
	end

	-- "mandatory" rewards (money, XP, items, etc that are rewarded regardless of any choices)
	if ( numQuestRewards > 0 or numQuestCurrencies > 0 or money > 0 or xp > 0 or honor > 0 or majorFactionRepRewards or hasChanceForQuestSessionBonusReward ) then
		-- receive text, will either say "You will receive" or "You will also receive"
		cql.Detail:AddFontStringToRewards((numQuestChoices>0 or #spellRewards>0 or playerTitle) and REWARD_ITEMS or REWARD_ITEMS_ONLY)

		-- experience
		if xp > 0 then
			cql.Detail:AddRewardButtonToRewards(questID,"reward",nil,"xp",xp,BreakUpLargeNumbers(xp),"Interface\\Icons\\XP_Icon")
		end
		-- money
		if money > 0 then
			cql.Detail:AddRewardButtonToRewards(questID,"reward",nil,"money",money,GetMoneyString(money),"Interface\\Icons\\inv_misc_coin_01")
		end
		-- skill points
		if skillPoints then
			cql.Detail:AddRewardButtonToRewards(questID,"reward",nil,"skill",format(BONUS_SKILLPOINTS_TOOLTIP, skillPoints, skillName),format(BONUS_SKILLPOINTS,skillName),skillIcon,nil,"\124cff19ff19+"..skillPoints)
		end
		-- item rewards
		for i=1,numQuestRewards do
			local name, texture, numItems, quality, isUsable, itemID = GetQuestLogRewardInfo(i)
			cql.Detail:AddRewardButtonToRewards(questID,"reward",i,"item",itemID,name,texture,quality,numItems)
		end
		-- currency (this is sometimes reputation too, like the Aiding the Accord quests)
		for i=1,numQuestCurrencies do
			local name, texture, amount, currencyID, quality = GetQuestLogRewardCurrencyInfo(i,questID)
			local info = C_CurrencyInfo.GetCurrencyContainerInfo(currencyID,amount)
			if info then -- if GetCurrencyContainerInfo changes any details
				cql.Detail:AddRewardButtonToRewards(questID,"reward",i,"currency",{currencyID=currencyID,amount=amount},info.name,info.icon,info.quality,info.displayAmount)
			else
				cql.Detail:AddRewardButtonToRewards(questID,"reward",i,"currency",{currencyID=currencyID,amount=amount},name,texture,quality,amount)
			end
		end
		-- reputation
		if majorFactionRepRewards then
			for i,rewardInfo in ipairs(majorFactionRepRewards) do
				local majorFactionData = C_MajorFactions.GetMajorFactionData(rewardInfo.factionID)
				local factionName = majorFactionData.name
				local rewardAmount = rewardInfo.rewardAmount
				cql.Detail:AddRewardButtonToRewards(questID,"reward",i,"reputation",{factionID=rewardInfo.factionID,factionName=factionName,rewardAmount=rewardAmount},format(QUEST_REPUTATION_REWARD_TITLE,factionName),format("Interface\\Icons\\UI_MajorFaction_%s",majorFactionData.textureKit),nil,AbbreviateNumbers(rewardAmount))
			end
		end
		-- warmode bonus
		if hasWarModeBonus and C_PVP.IsWarModeDesired() then
			local bonus = C_PvP.GetWarModeRewardBonus()
			cql.Detail:AddRewardButtonToRewards(questID,"reward",nil,"warmode",bonus,"\124cff19ff19"..WAR_MODE_BONUS,"Interface\\Icons\\UI_WARMODE",nil,"\124cff19ff19"..format(PLUS_PERCENT_FORMAT,bonus))
		end
		-- honor
		if honor > 0 then
			local icon = (UnitFactionGroup("player")==PLAYER_FACTION_GROUP[0]) and "Interface\\Icons\\PVPCurrency-Honor-Horde" or "Interface\\Icons\\PVPCurrency-Honor-Alliance"
			cql.Detail:AddRewardButtonToRewards(questID,"reward",nil,"currency",{currencyID=Constants.CurrencyConsts.HONOR_CURRENCY_ID,amount=honor},HONOR,icon,nil,honor)
		end
		-- bonus reward while in party sync
		if hasChanceForQuestSessionBonusReward then
			-- "Completing this quest while in Party Sync may reward:"
			cql.Detail:AddFontStringToRewards(QUEST_SESSION_BONUS_LOOT_REWARD_FRAME_TITLE)
			local QUEST_SESSION_BONUS_REWARD_ITEM_ID = 171305
			local QUEST_SESSION_BONUS_REWARD_ITEM_COUNT = 1
			local item = Item:CreateFromItemID(QUEST_SESSION_BONUS_REWARD_ITEM_ID)
			cql.Detail:AddRewardButtonToRewards(questID,"reward",nil,"questSessionBonusReward",item:GetItemID(),item:GetItemName(),item:GetItemIcon(),item:GetItemQuality(),QUEST_SESSION_BONUS_REWARD_ITEM_COUNT)
		end
	end

	rewardsFrame:SetHeight(rewardsFrame.totalHeight)
	rewardsFrame:Show()
	return rewardsFrame,rewardsFrame.lastAnchorElement
end
